# Filename: test_style.py
# pylint: disable=C0111,E1003,R0904,C0103,R0201,C0102

from km3pipe.testing import TestCase, patch
from km3pipe.style import get_style_path, ColourCycler, use

__author__ = "Tamas Gal"
__copyright__ = "Copyright 2016, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal"
__email__ = "tgal@km3net.de"
__status__ = "Development"


class TestStyle(TestCase):
    def test_get_style_path(self):
        gsp = get_style_path
        self.assertTrue(gsp("km3pipe").endswith("stylelib/km3pipe.mplstyle"))
        self.assertTrue(gsp("foo").endswith("/stylelib/foo.mplstyle"))
        self.assertTrue(gsp("bar").endswith("/stylelib/bar.mplstyle"))


class TestColourCycler(TestCase):
    def test_available(self):
        cc = ColourCycler()
        self.assertTrue("km3pipe" in cc.available)
        self.assertTrue("classic" in cc.available)

    def test_next(self):
        cc = ColourCycler()
        first = next(cc)
        second = next(cc)
        self.assertTrue(first != second)

    def test_next_a_few_times(self):
        cc = ColourCycler()
        last_colour = ""
        for i in range(100):
            current_colour = next(cc)
            self.assertTrue(last_colour != current_colour)
            last_colour = current_colour

    def test_raise_keyerror_if_style_not_available(self):
        with self.assertRaises(KeyError):
            cc = ColourCycler("foo")  # noqa


class TestStyles(TestCase):
    @patch("matplotlib.pyplot")
    def test_non_existent_style(self, plt_mock):
        use("non-existent")
        assert not plt_mock.style.use.called

    @patch("matplotlib.pyplot")
    def test_km3pipe(self, plt_mock):
        use("km3pipe")
        self._assert_plt_imported(plt_mock, "km3pipe.mplstyle")

    @patch("matplotlib.pyplot")
    def test_noargs_load_km3pipe_style(self, plt_mock):
        use()
        self._assert_plt_imported(plt_mock, "km3pipe.mplstyle")

    @patch("matplotlib.pyplot")
    def test_poster_style(self, plt_mock):
        use("poster")
        self._assert_plt_imported(plt_mock, "km3pipe-poster.mplstyle")

    @patch("matplotlib.pyplot")
    def test_notebook_style(self, plt_mock):
        use("notebook")
        self._assert_plt_imported(plt_mock, "km3pipe-notebook.mplstyle")

    @patch("matplotlib.pyplot")
    def test_talk_style(self, plt_mock):
        use("talk")
        self._assert_plt_imported(plt_mock, "km3pipe-talk.mplstyle")

    @patch("matplotlib.pyplot")
    def test_alba_style(self, plt_mock):
        use("alba")
        self._assert_plt_imported(plt_mock, "alba.mplstyle")

    @patch("matplotlib.pyplot")
    def test_jonas_style(self, plt_mock):
        use("jonas-phd")
        self._assert_plt_imported(plt_mock, "jonas-phd.mplstyle")

    @patch("matplotlib.pyplot")
    def test_jvs_style(self, plt_mock):
        use("jvs")
        self._assert_plt_imported(plt_mock, "jvs.mplstyle")

    @patch("matplotlib.pyplot")
    def test_moritz_style(self, plt_mock):
        use("moritz")
        self._assert_plt_imported(plt_mock, "moritz.mplstyle")

    @patch("matplotlib.pyplot")
    def test_serifs_style(self, plt_mock):
        use("serifs")
        self._assert_plt_imported(plt_mock, "serifs.mplstyle")

    @patch("matplotlib.pyplot")
    def test_import_alba(self, plt_mock):
        import km3pipe.style.alba

        self._assert_plt_imported(plt_mock, "alba.mplstyle")

    @patch("matplotlib.pyplot")
    def test_import_moritz(self, plt_mock):
        import km3pipe.style.moritz

        self._assert_plt_imported(plt_mock, "moritz.mplstyle")

    @patch("matplotlib.pyplot")
    def test_import_default(self, plt_mock):
        import km3pipe.style.default

        self._assert_plt_imported(plt_mock, "km3pipe.mplstyle")

    @patch("matplotlib.pyplot")
    def test_import_jonas_phd(self, plt_mock):
        import km3pipe.style.jonas_phd

        self._assert_plt_imported(plt_mock, "jonas-phd.mplstyle")

    @patch("matplotlib.pyplot")
    def test_import_km3pipe(self, plt_mock):
        import km3pipe.style.km3pipe

        self._assert_plt_imported(plt_mock, "km3pipe.mplstyle")

    @patch("matplotlib.pyplot")
    def test_import_km3pipe_notebook(self, plt_mock):
        import km3pipe.style.km3pipe_notebook

        self._assert_plt_imported(plt_mock, "km3pipe-notebook.mplstyle")

    @patch("matplotlib.pyplot")
    def test_import_km3pipe_poster(self, plt_mock):
        import km3pipe.style.km3pipe_poster

        self._assert_plt_imported(plt_mock, "km3pipe-poster.mplstyle")

    @patch("matplotlib.pyplot")
    def test_import_km3pipe_talk(self, plt_mock):
        import km3pipe.style.km3pipe_talk

        self._assert_plt_imported(plt_mock, "km3pipe-talk.mplstyle")

    def _assert_plt_imported(self, plt_mock, style_filename):
        args, kwargs = plt_mock.style.use.call_args_list[0]
        assert args[0].endswith(style_filename)



# Filename: cmd.py
"""
KM3Pipe command line utility.

Usage:
    km3pipe test
    km3pipe update [GIT_BRANCH]
    km3pipe createconf [--overwrite] [--dump]
    km3pipe retrieve DET_ID RUN [-i -o OUTFILE]
    km3pipe (-h | --help)
    km3pipe --version

Options:
    -h --help           Show this screen.
    -m                  Get the MC detector file (flips the sign of DET_ID).
    -c CALIBR_ID        Geometrical calibration ID (eg. A01466417)
    -i                  Use iRODS instead of xrootd to retrieve files.
    -o OUT              Output folder or filename.
    -t T0_SET           Time calibration ID (eg. A01466431)
    -s REGEX            Regular expression to filter the runsetup name/id.
    DET_ID              Detector ID (eg. D_ARCA001).
    DETECTOR            Detector (eg. ARCA).
    GIT_BRANCH          Git branch to pull (eg. develop).
    RUN                 Run number.

"""
import re
import sys
import subprocess
import os
from datetime import datetime
import time

import km3db

from . import version
from .tools import irods_path, xrootd_path
from .hardware import Detector

from km3pipe.logger import get_logger

log = get_logger(__name__)  # pylint: disable=C0103

__author__ = "Tamas Gal"
__copyright__ = "Copyright 2016, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal and Moritz Lotze"
__email__ = "tgal@km3net.de"
__status__ = "Development"

SPS_CACHE = "/sps/km3net/repo/data/cache"
KM3PIPE_GIT = "https://git.km3net.de/km3py/km3pipe.git"


def run_tests():
    print(
        "Running the km3pipe test suite. Make sure you have all the test "
        "dependencies and extras installed with\n\n"
        '    pip install "km3pipe[dev]"\n'
        '    pip install "km3pipe[extras]"\n'
    )

    import pytest
    import km3pipe

    pytest.main([os.path.dirname(km3pipe.__file__)])


def update_km3pipe(git_branch=""):
    if git_branch == "" or git_branch is None:
        git_branch = "master"
    os.system(f"pip install -U git+{KM3PIPE_GIT}@{git_branch}")


def retrieve(run_id, det_id, use_irods=False, out=None):
    """Retrieve run from iRODS for a given detector (O)ID"""
    det_id = int(det_id)

    if use_irods:
        rpath = irods_path(det_id, run_id)
        cmd = "iget -Pv"
    else:
        rpath = xrootd_path(det_id, run_id)
        cmd = "xrdcp"

    filename = os.path.basename(rpath)

    if out is not None and os.path.isdir(out):
        outfile = os.path.join(out, filename)
    elif out is None:
        outfile = filename
    else:
        outfile = out

    if os.path.exists(outfile):
        print("Output file '{}' already exists.".format(outfile))
        return

    cmd += " {0} {1}".format(rpath, outfile)

    if not km3db.core.on_whitelisted_host("lyon"):
        subprocess.call(cmd.split())
        return

    subfolder = os.path.join(*rpath.split("/")[-3:-1])
    cached_subfolder = os.path.join(SPS_CACHE, subfolder)
    cached_filepath = os.path.join(cached_subfolder, filename)
    lock_file = cached_filepath + ".in_progress"

    if os.path.exists(lock_file):
        print("File is already requested, waiting for the download to finish.")
        for _ in range(6 * 15):  # 15 minute timeout
            time.sleep(10)
            if not os.path.exists(lock_file):
                print("Done.")
                break
        else:
            print(
                "Timeout reached. Deleting the lock file and initiating a "
                "new download."
            )
            os.remove(lock_file)

    if not os.path.exists(cached_filepath):
        print("Downloading file to local SPS cache ({}).".format(SPS_CACHE))
        try:
            os.makedirs(cached_subfolder, exist_ok=True)
            subprocess.call(["touch", lock_file])
            subprocess.call(["chmod", "g+w", lock_file])
            subprocess.call(cmd.split())
            subprocess.call(["chmod", "g+w", outfile])
            subprocess.call(["cp", "-p", outfile, cached_filepath])
        except (OSError, KeyboardInterrupt) as e:
            print("\nAborting... {}".format(e))
            for f in [outfile, lock_file, cached_filepath]:
                subprocess.call(["rm", "-f", f])
            return
        finally:
            for f in [outfile, lock_file]:
                subprocess.call(["rm", "-f", f])

    subprocess.call(["ln", "-s", cached_filepath, outfile])


def main():
    from docopt import docopt

    args = docopt(
        __doc__,
        version="KM3Pipe {}".format(
            version,
        ),
    )

    if args["test"]:
        run_tests()

    if args["update"]:
        update_km3pipe(args["GIT_BRANCH"])

    if args["retrieve"]:
        retrieve(int(args["RUN"]), args["DET_ID"], use_irods=args["-i"], out=args["-o"])



# Filename: physics.py
# pylint: disable=locally-disabled
"""
Cherenkov photon parameters.

"""

import km3io
import numba
import numpy as np

import awkward as ak
from numba import njit

from thepipe import Module
from km3pipe.dataclasses import Table
from km3pipe.hardware import Detector
from km3pipe.logger import get_logger
from km3pipe.constants import (
    C_WATER,
    SIN_CHERENKOV,
    TAN_CHERENKOV,
    V_LIGHT_WATER,
    C_LIGHT,
)
import km3pipe.math
import km3pipe.extras

__author__ = "Zineb ALY"
__copyright__ = "Copyright 2020, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Zineb ALY"
__email__ = "zaly@km3net.de"
__status__ = "Development"

log = get_logger(__name__)


def cherenkov(calib_hits, track):
    """Compute parameters of Cherenkov photons emitted from a track and hitting a PMT.
    calib_hits is the table of calibrated hits of the track.

    Parameters
    ----------
    calib_hits : kp Table or a DataFrame or a numpy recarray or a dict.
        Table of calibrated hits with the following parameters:
            - pos_x.
            - pos_y.
            - pos_z.
            - dir_x.
            - dir_y.
            - dir_z.
    track : awkward.Array, DataFrame, numpy.recarray, km3pipe.Table or a dict.
        One track with the following parameters:
            - pos_x.
            - pos_y.
            - pos_z.
            - dir_x.
            - dir_y.
            - dir_z.
            - t.

    Returns
    -------
    ndarray
        a structured array of the physics parameters of Cherenkov photons:
            - d_photon_closest: the closest distance between the PMT and the track
            (it is perpendicular to the track).
            - d_photon: distance traveled by the photon from the track to the PMT.
            - d_track: distance along the track where the photon was emitted.
            - t_photon: time of photon travel in [s].
            - cos_photon_PMT: cos angle of impact of photon with respect to the PMT direction:
            - dir_x_photon, dir_y_photon, dir_z_photon: photon directions.
    """

    pd = km3pipe.extras.pandas()

    if isinstance(calib_hits, (dict, pd.DataFrame, np.ndarray, Table, ak.Record)):
        calib_pos = np.array(
            [calib_hits["pos_x"], calib_hits["pos_y"], calib_hits["pos_z"]]
        ).T
        calib_dir = np.array(
            [calib_hits["dir_x"], calib_hits["dir_y"], calib_hits["dir_z"]]
        ).T

    if isinstance(track, (dict, pd.core.series.Series, pd.DataFrame)):
        track_pos = np.array([track["pos_x"], track["pos_y"], track["pos_z"]]).T
        track_dir = np.array([track["dir_x"], track["dir_y"], track["dir_z"]]).T
        track_t = track["t"]

    if isinstance(track, (Table, np.ndarray)):
        track_pos = np.array([track["pos_x"], track["pos_y"], track["pos_z"]]).reshape(
            3,
        )
        track_dir = np.array([track["dir_x"], track["dir_y"], track["dir_z"]]).reshape(
            3,
        )
        track_t = track["t"][0]

    if isinstance(track, (ak.Array, ak.Record, km3io.rootio.Branch)):
        track_pos = np.array([track.pos_x, track.pos_y, track.pos_z]).T
        track_dir = np.array([track.dir_x, track.dir_y, track.dir_z]).T
        track_t = track.t

    out = _cherenkov(calib_pos, calib_dir, track_pos, track_dir, track_t)

    return out.view(
        dtype=[
            ("d_photon_closest", "<f8"),
            ("d_photon", "<f8"),
            ("d_track", "<f8"),
            ("t_photon", "<f8"),
            ("cos_photon_PMT", "<f8"),
            ("dir_x_photon", "<f8"),
            ("dir_y_photon", "<f8"),
            ("dir_z_photon", "<f8"),
        ]
    ).reshape(
        len(out),
    )


@njit
def _cherenkov(calib_pos, calib_dir, track_pos, track_dir, track_t):
    """calculate Cherenkov photons parameters"""
    rows = len(calib_pos)
    out = np.zeros((rows, 8))

    for i in range(rows):
        # (vector PMT) - (vector track position)
        V = calib_pos[i] - track_pos
        L = np.sum(V * track_dir)
        out[i, 0] = np.sqrt(np.sum(V * V) - L * L)  # d_photon_closest
        out[i, 1] = out[i][0] / SIN_CHERENKOV  # d_photon
        out[i, 2] = L - out[i][0] / TAN_CHERENKOV  # d_track
        out[i, 3] = (
            track_t + out[i][2] / C_LIGHT + out[i][1] / V_LIGHT_WATER
        )  # t_photon
        V_photon = V - (out[i][2] * track_dir)  # photon position
        norm = np.sqrt(np.sum(V_photon * V_photon))
        out[i, 5:8] = V_photon / norm  # photon direction

        # cos angle of impact of photon with respect to the PMT direction
        out[i, 4] = np.sum(out[i][5:8] * calib_dir[i])
    return out


def _get_closest(track_pos, track_dir, meanDU_pos, meanDU_dir):
    """calculate the distance of closest approach"""
    # direction track to the mean DU, assumes vertical DU:
    cross = np.cross(track_dir, meanDU_dir)
    dir_to_DU = cross / np.linalg.norm(cross, axis=0)  # normalized vector

    # vector track position to mean DU position, perpendicular to track.
    V = track_pos - meanDU_pos
    d_closest = abs(np.dot(V, dir_to_DU))

    # find z-coordinate of the point of closest approach
    # 1) - vector track position to mean DU position, assuming vertical DU.
    V[2] = 0.0
    V_norm = np.linalg.norm(V, axis=0)

    # 2) distance from the track origin to point of closest approach along the track
    d = np.sqrt(V_norm ** 2 - d_closest ** 2)
    # norm of the track direction along (x,y)
    xy_norm = np.linalg.norm(track_dir[0:2], axis=0)

    # 3) track_pos_Z + distance (along z axis) to closest approach along the track
    z_closest = track_pos[2] + d * (track_dir[2] / xy_norm)

    return d_closest, z_closest


def get_closest(track, du_pos):
    """calculate the distance of closest approach (d_closest) and its coordinate along the z axis (z_closest).
    These calculations aLWAYS assume vertical DU.

    Parameters
    ----------
    track : awkward.Array, DataFrame, numpy.recarray, km3pipe.Table or a dict.
        One track with the following parameters:
            - pos_x.
            - pos_y.
            - pos_z.
            - dir_x.
            - dir_y.
            - dir_z.
    du_pos : a DataFrame or a numpy recarray or a km3pipe Table or a dict.
        du_pos vector with the following information:
            - pos_x.
            - pos_y.
            - pos_z.

    Returns
    -------
    tuple
        (d_closest, z_closest).
    """
    pd = km3pipe.extras.pandas()

    if isinstance(
        du_pos, (dict, pd.core.series.Series, pd.DataFrame, Table, np.ndarray)
    ):
        meanDU_pos = np.array(
            [du_pos["pos_x"], du_pos["pos_y"], du_pos["pos_z"]]
        ).reshape(
            3,
        )
        meanDU_dir = np.array([0, 0, 1])  # assumes vertical DU

    if isinstance(
        track, (dict, pd.core.series.Series, pd.DataFrame, Table, np.ndarray)
    ):
        track_pos = np.array([track["pos_x"], track["pos_y"], track["pos_z"]]).reshape(
            3,
        )
        track_dir = np.array([track["dir_x"], track["dir_y"], track["dir_z"]]).reshape(
            3,
        )

    if isinstance(track, ak.Array):
        track_pos = np.array([track.pos_x, track.pos_y, track.pos_z])
        track_dir = np.array([track.dir_x, track.dir_y, track.dir_z])

    return _get_closest(track_pos, track_dir, meanDU_pos, meanDU_dir)


def cut4d(point4d, tmin, tmax, rmin, rmax, items, c_water=C_WATER):
    """Select items with a certain time residual and
    within a certain radius around a given 4D point.

    Parameters
    -----------
    point4d: array of shape(x, y, z, t)
        central point of the selection
    tmin: float
        minimum tres of the sphere selection
    tmax: float
        maximum tres of the sphere selection
    rmin: float
        minimum radius of the sphere selection
    rmax: float
        maximum radius of the sphere selection
    items: iterable with pos_[xyz]-attributed items
        the items to cut on

    Returns
    --------
    iterable with pos_[xyz]-attributed items
        items which survived the cut.
    """
    point_array = point4d

    if all(hasattr(point4d, "pos_" + q) for q in "xyz"):
        point_array = np.array(
            [point4d[0].pos_x, point4d[0].pos_y, point4d[0].pos_z, point4d[0].t]
        )

    dt = items.time - point_array[3]

    items_pos = np.array([items.pos_x, items.pos_y, items.pos_z]).T

    distances = km3pipe.math.dist(
        np.array([point_array[0], point_array[1], point_array[2]]), items_pos, axis=1
    )

    tres = dt - (distances / C_WATER)

    mask = (tres >= tmin) & (tres <= tmax) & (distances >= rmin) & (distances <= rmax)
    selected_items = items[mask]

    return selected_items



# Filename: daq.py
# pylint: disable=R0903
"""
Pumps for the DAQ data formats.

"""

from collections import namedtuple
from io import BytesIO
import json
import math
import struct
from struct import unpack
import time
import pprint
from urllib.request import urlopen, URLError

import numpy as np

from thepipe import Module, Blob
from km3pipe.dataclasses import Table
from km3pipe.sys import ignored
from km3pipe.logger import get_logger, get_printer

__author__ = "Tamas Gal"
__copyright__ = "Copyright 2016, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal"
__email__ = "tgal@km3net.de"
__status__ = "Development"

log = get_logger(__name__)  # pylint: disable=C0103

DATA_TYPES = {
    101: "DAQSuperFrame",
    201: "DAQSummaryFrame",
    # Using the same class for all timeslices since they are structurally
    # identical (until now)
    1001: "DAQTimeslice",  # Type of erroneous timeslice data
    1002: "DAQTimeslice",  # L0
    1003: "DAQTimeslice",  # L1
    1004: "DAQTimeslice",  # L2
    1005: "DAQTimeslice",  # SN
    2001: "DAQSummaryslice",
    10001: "DAQEvent",
}
MINIMAL_RATE_HZ = 2.0e3
MAXIMAL_RATE_HZ = 2.0e6


class TimesliceParser(Module):
    """Preliminary parser for DAQTimeslice"""

    def configure(self):
        self.legacy = self.get("legacy", default=False)

    def _get_raw_data(self, blob):
        if "CHPrefix" in blob:
            if not str(blob["CHPrefix"].tag).startswith("IO_TS"):
                log.info("Not an IO_TS* blob")
                return blob
            return BytesIO(blob["CHData"])
        if "FileIO" in blob:
            return blob["FileIO"]
        if "RawBytes" in blob:
            return BytesIO(blob["RawBytes"])

    def process(self, blob):
        data = self._get_raw_data(blob)
        if data is None:
            return blob
        try:
            ts_info, ts_frameinfos, ts_hits = self._parse_timeslice(data)
        except struct.error:
            log.error("Could not parse Timeslice")
            log.error(blob.keys())
        else:
            blob["TSHits"] = ts_hits
            blob["TimesliceInfo"] = ts_info
            blob["TimesliceFrameInfos"] = ts_frameinfos
            return blob

    def _parse_timeslice(self, data):
        tsl_size, datatype = unpack("<ii", data.read(8))
        if not self.legacy:
            version = unpack("<h", data.read(2))[0]
            if version != 1:
                raise ValueError(
                    "Unsupported DAQTimeslice version ({}) or legacy DAQ. "
                    "Make sure Jpp v13+ is used or pass 'legacy=True' "
                    "to {}.".format(version, self.__class__.__name__)
                )
        det_id, run, sqnr = unpack("<iii", data.read(12))
        timestamp, ns_ticks, n_frames = unpack("<iii", data.read(12))

        ts_info = Table.from_template(
            {
                "frame_index": sqnr,
                "slice_id": 0,
                "timestamp": timestamp,
                "nanoseconds": ns_ticks * 16,
                "n_frames": n_frames,
            },
            "TimesliceInfo",
        )
        ts_frameinfos = {}

        _dom_ids = []
        _channel_ids = []
        _times = []
        _tots = []
        for _ in range(n_frames):
            frame_size, datatype = unpack("<ii", data.read(8))
            det_id, run, sqnr = unpack("<iii", data.read(12))
            timestamp, ns_ticks, dom_id = unpack("<iii", data.read(12))
            dataqueue_status = unpack("<i", data.read(4))[0]
            dom_status = unpack("<iiii", data.read(4 * 4))
            n_hits = unpack("<i", data.read(4))[0]
            ts_frameinfos[dom_id] = Table.from_template(
                {
                    "det_id": det_id,
                    "run_id": run,
                    "sqnr": sqnr,
                    "timestamp": timestamp,
                    "nanoseconds": ns_ticks * 16,
                    "dom_id": dom_id,
                    "dataqueue_status": dataqueue_status,
                    "dom_status": dom_status,
                    "n_hits": n_hits,
                },
                "TimesliceFrameInfo",
            )
            for j in range(n_hits):
                hit = unpack("!BlB", data.read(6))
                _dom_ids.append(dom_id)
                _channel_ids.append(hit[0])
                _times.append(hit[1])
                _tots.append(hit[2])

        ts_hits = Table(
            {
                "channel_id": np.array(_channel_ids),
                "dom_id": np.array(_dom_ids),
                "time": np.array(_times),
                "tot": np.array(_tots),
            },
            name="TimesliceHits",
            h5loc="/timeslice_hits",
            split_h5=True,
        )
        return ts_info, ts_frameinfos, ts_hits


class RePump(Module):
    """A pump for binary DAQ files.


    This pump can be used to replay raw dumps e.g. created with the ``daqsample``
    tool. It creates the same structures as the ``kp.io.ch.CHPump`` and thus
    suited to test online processing pipelines with offline files.

    """

    def configure(self):
        self.filename = self.require("filename")
        self.fobj = open(self.filename, "rb")

    def process(self, blob):
        try:
            length, data_type = unpack("<ii", self.fobj.read(8))
            self.fobj.seek(-8, 1)
        except struct.error:
            raise StopIteration
        data = self.fobj.read(length)
        blob["RawBytes"] = data
        return blob

    def finish(self):
        self.fobj.close()


class DAQPump(Module):
    """A pump for binary DAQ files. Deprecated!"""

    def configure(self):
        self.filename = self.require("filename")
        self.legacy = self.get("legacy", default=False)
        self.frame_positions = []
        self.index = 0

        self.blob_file = self.open_file(self.filename)
        self.determine_frame_positions()

    def next_blob(self):
        """Get the next frame from file"""
        blob_file = self.blob_file
        try:
            preamble = DAQPreamble(file_obj=blob_file)
        except struct.error:
            raise StopIteration

        try:
            data_type = DATA_TYPES[preamble.data_type]
        except KeyError:
            log.error("Unknown datatype: {0}".format(preamble.data_type))
            data_type = "Unknown"

        blob = Blob()
        blob[data_type] = None
        blob["DAQPreamble"] = preamble

        if data_type == "DAQSummaryslice":
            daq_frame = DAQSummaryslice(blob_file, legacy=self.legacy)
            blob[data_type] = daq_frame
            blob["DAQHeader"] = daq_frame.header
        elif data_type == "DAQEvent":
            daq_frame = DAQEvent(blob_file, legacy=self.legacy)
            blob[data_type] = daq_frame
            blob["DAQHeader"] = daq_frame.header
        else:
            log.warning(
                "Skipping DAQ frame with data type code '{0}'.".format(
                    preamble.data_type
                )
            )
            blob_file.seek(preamble.length - DAQPreamble.size, 1)

        return blob

    def seek_to_frame(self, index):
        """Move file pointer to the frame with given index."""
        pointer_position = self.frame_positions[index]
        self.blob_file.seek(pointer_position, 0)

    def get_blob(self, index):
        """Return blob at given index."""
        self.seek_to_frame(index)
        return self.next_blob()

    def determine_frame_positions(self):
        """Record the file pointer position of each frame"""
        self.blob_file.seek(0, 0)
        with ignored(struct.error):
            while True:
                pointer_position = self.blob_file.tell()
                length = struct.unpack("<i", self.blob_file.read(4))[0]
                self.blob_file.seek(length - 4, 1)
                self.frame_positions.append(pointer_position)
        self.blob_file.seek(0, 0)
        log.info("Found {0} frames.".format(len(self.frame_positions)))

    def process(self, blob):
        """Pump the next blob to the modules"""
        return self.next_blob()

    def finish(self):
        """Clean everything up"""
        self.blob_file.close()

    def __len__(self):
        if not self.frame_positions:
            self.determine_frame_positions()
        return len(self.frame_positions)

    def __iter__(self):
        return self

    def __next__(self):
        try:
            blob = self.get_blob(self.index)
        except IndexError:
            self.index = 0
            raise StopIteration
        self.index += 1
        return blob

    def __getitem__(self, index):
        if isinstance(index, int):
            return self.get_blob(index)
        elif isinstance(index, slice):
            return self._slice_generator(index)
        else:
            raise TypeError("index must be int or slice")

    def _slice_generator(self, index):
        """A simple slice generator for iterations"""
        start, stop, step = index.indices(len(self))
        for i in range(start, stop, step):
            yield self.get_blob(i)


class DAQProcessor(Module):
    def configure(self):
        self.legacy = self.get("legacy", default=False)
        self.index = 0
        self.event_id = 0

    def process(self, blob):
        tag = str(blob["CHPrefix"].tag)
        data = blob["CHData"]

        processor = None

        if tag == "IO_EVT":
            processor = self.process_event
        if tag == "IO_SUM":
            processor = self.process_summaryslice
        if tag == "IO_OLINE":
            processor = self.process_online_reco

        if processor is None:
            self.log.error("Unsupported tag: %s", tag)
            return

        try:
            processor(data, blob)
        except (struct.error, ValueError) as e:
            self.log.error("Corrupt data received. Skipping...\n" "Error: %s", e)
            return

        return blob

    def process_event(self, data, blob):
        data_io = BytesIO(data)
        preamble = DAQPreamble(file_obj=data_io)  # noqa
        event = DAQEvent(file_obj=data_io, legacy=self.legacy)
        header = event.header

        event_info = Table.from_template(
            {
                "det_id": header.det_id,
                # 'frame_index': self.index,  # header.time_slice,
                "frame_index": header.time_slice,
                "livetime_sec": 0,
                "mc_id": 0,
                "mc_t": 0,
                "n_events_gen": 0,
                "n_files_gen": 0,
                "overlays": event.overlays,
                "trigger_counter": event.trigger_counter,
                "trigger_mask": event.trigger_mask,
                "utc_nanoseconds": header.ticks * 16,
                "utc_seconds": header.time_stamp,
                "weight_w1": 0,
                "weight_w2": 0,
                "weight_w3": 0,  # MC weights
                "run_id": header.run,  # run id
                "group_id": self.event_id,
            },
            "EventInfo",
        )
        blob["EventInfo"] = event_info

        self.event_id += 1
        self.index += 1

        hits = event.snapshot_hits
        n_hits = event.n_snapshot_hits
        if n_hits == 0:
            self.log.warning("No hits found in event.")
            return

        # This might not be needed
        triggereds = np.zeros(n_hits)
        triggered_map = {}
        for thit in event.triggered_hits:
            # TODO: switch to thit.trigger_mask instead of True!
            triggered_map[(thit.dom_id, thit.channel_id, thit.time, thit.tot)] = True
        for idx, hit in enumerate(hits):
            triggereds[idx] = tuple(hit) in triggered_map

        hit_series = Table.from_template(
            {
                "channel_id": hits.channel_id,
                "dom_id": hits.dom_id,
                "time": hits.time,
                "tot": hits.tot,
                "triggered": triggereds,  # TODO: switch to trigger_mask!
                "group_id": self.event_id,
            },
            "Hits",
        )

        blob["Hits"] = hit_series

    def process_summaryslice(self, data, blob):
        data_io = BytesIO(data)
        preamble = DAQPreamble(file_obj=data_io)  # noqa
        summaryslice = DAQSummaryslice(file_obj=data_io, legacy=self.legacy)
        blob["RawSummaryslice"] = summaryslice

    def process_online_reco(self, data, blob):
        data_io = BytesIO(data)
        preamble = DAQPreamble(file_obj=data_io)  # noqa
        _data = unpack("<iiiQI", data_io.read(4 + 4 + 4 + 8 + 4))
        det_id, run_id, frame_index, trigger_counter, utc_seconds = _data
        shower_reco = unpack("9d", data_io.read(9 * 8))
        shower_meta = unpack("3i", data_io.read(12))
        track_reco = unpack("9d", data_io.read(9 * 8))
        track_meta = unpack("3i", data_io.read(12))
        print(
            "Shower: x/y/z/dx/dy/dz/E/Q/t (type/status/ndf): ", shower_reco, shower_meta
        )
        print("Track: x/y/z/dx/dy/dz/E/Q/t (type/status/ndf): ", track_reco, track_meta)
        blob["ReconstructionInfo"] = Table(
            {
                "det_id": det_id,
                "run_id": run_id,
                "frame_index": frame_index,
                "trigger_counter": trigger_counter,
                "utc_seconds": utc_seconds,
            },
            h5loc="reco",
            split_h5=True,
            name="Reconstructions",
        )
        args = track_reco + track_meta
        blob["RecoTrack"] = RecoTrack(*args)
        args = shower_reco + shower_meta
        blob["RecoShower"] = RecoShower(*args)


RecoTrack = namedtuple("RecoTrack", "x y z dx dy dz E Q t type status ndf")
RecoShower = namedtuple("RecoShower", "x y z dx dy dz E Q t type status ndf")


class DAQPreamble(object):
    """Wrapper for the JDAQPreamble binary format.

    Parameters
    ----------
    byte_data : bytes (optional)
        The binary file, where the file pointer is at the beginning of the header.
    file_obj : file (optional)
        The binary file, where the file pointer is at the beginning of the header.

    Attributes
    ----------
    size : int
        The size of the original DAQ byte representation.
    data_type : int
        The data type of the following frame. The coding is stored in the
        ``DATA_TYPES``.

    """

    size = 8

    def __init__(self, byte_data=None, file_obj=None):
        self.length = None
        self.data_type = None
        if byte_data:
            self._parse_byte_data(byte_data)
        if file_obj:
            self._parse_file(file_obj)

    def _parse_byte_data(self, byte_data):
        """Extract the values from byte string."""
        self.length, self.data_type = unpack("<ii", byte_data[: self.size])

    def _parse_file(self, file_obj):
        """Directly read from file handler.

        Note that this will move the file pointer.

        """
        byte_data = file_obj.read(self.size)
        self._parse_byte_data(byte_data)

    def __repr__(self):
        description = "Length: {0}\nDataType: {1}".format(self.length, self.data_type)
        return description


class DAQHeader(object):
    """Wrapper for the JDAQHeader binary format.

    Parameters
    ----------
    byte_data : bytes (optional)
        The binary file, where the file pointer is at the beginning of the header.
    file_obj : file (optional)
        The binary file, where the file pointer is at the beginning of the header.

    Attributes
    ----------
    size : int
        The size of the original DAQ byte representation.

    """

    size = 20

    def __init__(self, byte_data=None, file_obj=None):
        self.run = None
        self.time_slice = None
        self.time_stamp = None
        if byte_data:
            self._parse_byte_data(byte_data)
        if file_obj:
            self._parse_file(file_obj)

    def _parse_byte_data(self, byte_data):
        """Extract the values from byte string."""
        chunks = unpack("<iiiii", byte_data[: self.size])
        det_id, run, time_slice, time_stamp, ticks = chunks
        self.det_id = det_id
        self.run = run
        self.time_slice = time_slice
        self.time_stamp = time_stamp
        self.ticks = ticks

    def _parse_file(self, file_obj):
        """Directly read from file handler.

        Note:
          This will move the file pointer.

        """
        byte_data = file_obj.read(self.size)
        self._parse_byte_data(byte_data)

    def __repr__(self):
        description = "Run: {0}\nTime slice: {1}\nTime stamp: {2} ({3})".format(
            self.run, self.time_slice, self.time_stamp, self.ticks
        )
        return description


class DAQSummaryslice(object):
    """Wrapper for the JDAQSummarySlice binary format.

    Parameters
    ----------
    file_obj : file (optional)
        The binary file, where the file pointer is at the beginning of the header.

    Attributes
    ----------
    n_summary_frames : int
        The number of summary frames.
    summary_frames : dict
        The PMT rates for each DOM. The key is the DOM identifier and the
        corresponding value is a sorted list of PMT rates in [Hz].
    dom_rates : dict
        The overall DOM rate for each DOM.

    """

    def __init__(self, file_obj, legacy=False):
        if not legacy:
            version = unpack("<h", file_obj.read(2))[0]
            if version != 6:
                raise ValueError(
                    "Unsupported {} version ({}) or legacy DAQ. "
                    "Make sure Jpp v13+ is used or pass 'legacy=True' "
                    "to the init.".format(self.__class__.__name__, version)
                )
        self.header = DAQHeader(file_obj=file_obj)
        self.n_summary_frames = unpack("<i", file_obj.read(4))[0]
        self.summary_frames = {}
        self.dq_status = {}
        self.dom_status = {}
        self.dom_rates = {}

        self._parse_summary_frames(file_obj)

    def _parse_summary_frames(self, file_obj):
        """Iterate through the byte data and fill the summary_frames"""
        for _ in range(self.n_summary_frames):
            dom_id = unpack("<i", file_obj.read(4))[0]
            dq_status = file_obj.read(4)  # probably dom status? # noqa
            dom_status = unpack("<iiii", file_obj.read(16))
            raw_rates = unpack("b" * 31, file_obj.read(31))
            pmt_rates = [self._get_rate(value) for value in raw_rates]
            self.summary_frames[dom_id] = pmt_rates
            self.dq_status[dom_id] = dq_status
            self.dom_status[dom_id] = dom_status
            self.dom_rates[dom_id] = np.sum(pmt_rates)

    def _get_rate(self, value):
        """Return the rate in Hz from the short int value"""
        if value == 0:
            return 0
        else:
            return MINIMAL_RATE_HZ * math.exp(value * self._get_factor())

    def _get_factor(self):
        return math.log(MAXIMAL_RATE_HZ / MINIMAL_RATE_HZ) / 255


class DAQEvent(object):
    """Wrapper for the JDAQEvent binary format.

    Parameters
    ----------
    file_obj : file
        The binary file, where the file pointer is at the beginning of the header.


    Attributes
    ----------
    trigger_counter : int
        Incremental identifier of the occurred trigger.
    trigger_mask : int
        The trigger type(s) satisfied.
    overlays : int
        Number of merged events.
    n_triggered_hits : int
        Number of hits satisfying the trigger conditions.
    n_snapshot_hits : int
        Number of snapshot hits.
    triggered_hits : np.recarray
        Array of triggered hits (fields: dom_id, pmt_id, tdc_time, tot, trigger_mask)
    snapshot_hits : np.recarray
        A list of snapshot hits (fields: dom_id, pmt_id, tdc_time, tot)

    """

    triggered_hits_dt = np.dtype(
        [
            ("dom_id", "<i"),
            ("channel_id", np.uint8),
            ("time", ">I"),
            ("tot", np.uint8),
            ("trigger_mask", "<Q"),
        ]
    )
    snapshot_hits_dt = np.dtype(
        [
            ("dom_id", "<i"),
            ("channel_id", np.uint8),
            ("time", ">I"),
            ("tot", np.uint8),
        ]
    )

    def __init__(self, file_obj, legacy=False):
        if not legacy:
            version = unpack("<h", file_obj.read(2))[0]
            if version != 4:
                raise ValueError(
                    "Unsupported {} version ({}) or legacy DAQ. "
                    "Make sure Jpp v13+ is used or pass 'legacy=True' "
                    "to the init.".format(self.__class__.__name__, version)
                )
        self.header = DAQHeader(file_obj=file_obj)
        self.trigger_counter = unpack("<Q", file_obj.read(8))[0]
        self.trigger_mask = unpack("<Q", file_obj.read(8))[0]
        self.overlays = unpack("<i", file_obj.read(4))[0]

        self.n_triggered_hits = unpack("<i", file_obj.read(4))[0]
        self.triggered_hits = self._parse_triggered_hits(file_obj)

        self.n_snapshot_hits = unpack("<i", file_obj.read(4))[0]
        self.snapshot_hits = self._parse_snapshot_hits(file_obj)

    def _parse_triggered_hits(self, file_obj):
        """Parse and store triggered hits."""
        raw_data = file_obj.read(
            self.triggered_hits_dt.itemsize * self.n_triggered_hits
        )
        return np.frombuffer(raw_data, self.triggered_hits_dt).view(np.recarray)

    def _parse_snapshot_hits(self, file_obj):
        """Parse and store snapshot hits."""
        raw_data = file_obj.read(self.snapshot_hits_dt.itemsize * self.n_snapshot_hits)
        return np.frombuffer(raw_data, self.snapshot_hits_dt).view(np.recarray)

    def __repr__(self):
        string = "\n".join(
            (
                " Number of triggered hits: " + str(self.n_triggered_hits),
                " Number of snapshot hits: " + str(self.n_snapshot_hits),
            )
        )
        string += "\nTriggered hits:\n"
        string += pprint.pformat(self.triggered_hits)
        string += "\nSnapshot hits:\n"
        string += pprint.pformat(self.snapshot_hits)
        return string


class TMCHData(object):
    """Monitoring Channel data."""

    def __init__(self, file_obj, version=None):
        f = file_obj

        data_type = f.read(4)
        if data_type != b"TMCH":
            raise ValueError("Invalid datatype: {0}".format(data_type))

        self.run = unpack(">I", f.read(4))[0]
        self.udp_sequence_number = unpack(">I", f.read(4))[0]
        self.utc_seconds = unpack(">I", f.read(4))[0]
        self.nanoseconds = unpack(">I", f.read(4))[0] * 16
        self.dom_id = unpack(">I", f.read(4))[0]
        self.dom_status_0 = unpack(">I", f.read(4))[0]
        self.dom_status_1 = unpack(">I", f.read(4))[0]
        self.dom_status_2 = unpack(">I", f.read(4))[0]
        self.dom_status_3 = unpack(">I", f.read(4))[0]
        self.pmt_rates = [r * 10.0 for r in unpack(">" + 31 * "I", f.read(31 * 4))]
        self.hrvbmp = unpack(">I", f.read(4))[0]
        self.flags = unpack(">I", f.read(4))[0]
        # flags:
        # bit 0: AHRS valid
        if version is None:
            # bit 3-1: structure version
            #          000 - 1, 001 - 2, 010 - unused, 011 - 3
            self.version = int(bin((self.flags >> 1) & 7), 2) + 1
        else:
            self.version = version
        self.yaw, self.pitch, self.roll = unpack(">fff", f.read(12))
        self.A = unpack(">fff", f.read(12))  # AHRS: Ax, Ay, Az
        self.G = unpack(">fff", f.read(12))  # AHRS: Gx, Gy, Gz
        self.H = unpack(">fff", f.read(12))  # AHRS: Hx, Hy, Hz
        self.temp = unpack(">H", f.read(2))[0] / 100.0
        self.humidity = unpack(">H", f.read(2))[0] / 100.0
        self.tdcfull = unpack(">I", f.read(4))[0]
        self.aesfull = unpack(">I", f.read(4))[0]
        self.flushc = unpack(">I", f.read(4))[0]

        if self.version >= 2:
            self.ts_duration_ms = unpack(">I", f.read(4))[0]
        if self.version >= 3:
            self.tdc_supertime_fifo_size = unpack(">H", f.read(2))[0]
            self.aes_supertime_fifo_size = unpack(">H", f.read(2))[0]

    def __str__(self):
        return str(vars(self))

    def __repr__(self):
        return self.__str__()


class TMCHRepump(Module):
    """Takes a IO_MONIT raw dump and replays it."""

    def configure(self):
        filename = self.require("filename")
        self.format_version = self.get("format_version", default=None)
        self.fobj = open(filename, "rb")
        self.blobs = self.blob_generator()

    def process(self, blob):
        return next(self.blobs)

    def blob_generator(self):
        while True:
            blob = Blob()
            datatype = self.fobj.read(4)
            if len(datatype) == 0:
                return
            if datatype == b"TMCH":
                self.fobj.seek(-4, 1)
                blob["TMCHData"] = TMCHData(self.fobj, version=self.format_version)
                yield blob

    def finish(self):
        self.fobj.close()

    def __iter__(self):
        return self

    def __next__(self):
        return next(self.blobs)


class DMMonitor(object):
    """A class which provides access to the Detector Manager parameters.

    Examples
    --------
    >>> import km3pipe as kp
    >>> dmm = kp.io.daq.DMMonitor('192.168.0.120', base='clb/outparams')
    >>> session = dmm.start_session('test', ['wr_mu/1/0','wr_mu/1/1'])
    >>> for values in session:
            print(values)

    """

    def __init__(self, host, port=1302, base=""):
        self._host = host
        self._port = port
        self._base = base
        self._url = "http://{}:{}/mon/{}".format(self._host, self._port, self._base)
        self._available_parameters = []
        self.log = get_logger(self.__class__.__name__)
        self.cprint = get_printer(self.__class__.__name__)

    @property
    def available_parameters(self):
        if not self._available_parameters:
            self._get_available_parameters()
        return self._available_parameters

    def _get_available_parameters(self):
        self._available_parameters = json.loads(urlopen(self._url).read())

    def get(self, path):
        return json.loads(
            urlopen(
                "http://{}:{}/mon/{}/{}".format(
                    self._host, self._port, self._base, path
                )
            ).read()
        )

    def start_session(self, name, paths, interval=10):
        self.cprint("Starting session '{}'".format(name))
        ret = urlopen(
            "http://{}:{}/monshortdef?name={}&paths={}".format(
                self._host,
                self._port,
                name,
                ",".join(["/mon/{}/{}".format(self._base, p) for p in paths]),
            )
        ).read()
        if ret != b"OK":
            self.log.error("Could not start session")
            return []
        return self._session(name, interval)

    def _session(self, name, interval):
        url = "http://{}:{}/monshort/{}".format(self._host, self._port, name)
        while True:
            try:
                yield json.loads(urlopen(url).read())
            except URLError as e:
                self.log.error(
                    "Error when trying to connect to the DM: %s\n"
                    "Retry in %d seconds..." % (e, interval)
                )
            time.sleep(interval)



#!/usr/bin/env python
# coding: utf-8 -*-
"""

==================
PMT Time Slewing
==================

Show different variants of PMT time slewing calculations.

Time slewing corrects the hit time due to different rise times of the
PMT signals depending on the number of photo electrons.
The reference point is at 26.4ns and hits with a different ToT values
are corrected to refer to comparable arrival times.
The time slewing is subtracted from the measured hit time, in contrast
to the time calibration (t0), which is added.

The time slewing correction is automatically applied in `km3pipe` when
using `kp.calib.Calibration().apply()`, however it can be turned off
by providing `correct_slewing=False` and also the variant can be picked
by `slewing_variant=X`.

Variant 3 is currently (as of 2020-10-16) also used in Jpp.

"""

# Author: Tamas Gal <tgal@km3net.de>
# License: BSD-3

import km3pipe as kp
import numpy as np
import matplotlib.pyplot as plt

kp.style.use()

tots = np.arange(256)

##############################
# The `kp.cali.slew()` function can be used to calculate the slew.
# It takes a single ToT or an array of ToTs and optionally a variant.
# Here is the docstring of the function:

help(kp.calib.slew)

##############################
# Calculating the slew for all variants:

slews = {variant: kp.calib.slew(tots, variant=variant) for variant in (1, 2, 3)}

fig, ax = plt.subplots()

for variant, slew in slews.items():
    ax.plot(tots, slew, label=f"Variant {variant}")

ax.set_xlabel("ToT / ns")
ax.set_ylabel("time slewing / ns")
ax.legend()

fig.tight_layout()
plt.show()



#!/usr/bin/env python
# Filename: jpp.py
# pylint: disable=
"""
Pump for the jpp file read through aanet interface.

"""

from thepipe import Module, Blob
from km3pipe.controlhost import Client
from km3pipe.time import Cuckoo
from km3pipe.logger import get_logger
import threading
import socket
import time
import numpy as np
from collections import deque
from queue import Queue, Empty

__author__ = "Tamas Gal"
__copyright__ = "Copyright 2016, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal"
__email__ = "tgal@km3net.de"
__status__ = "Development"

log = get_logger(__name__)  # pylint: disable=C0103


class CHPump(Module):
    """A pump for ControlHost data."""

    def configure(self):
        self.host = self.get("host") or "127.0.0.1"
        self.port = self.get("port") or 5553
        self.tags = self.get("tags") or "MSG"
        self.timeout = self.get("timeout") or 60 * 60 * 24
        self.max_queue = self.get("max_queue") or 50
        self.key_for_data = self.get("key_for_data") or "CHData"
        self.key_for_prefix = self.get("key_for_prefix") or "CHPrefix"
        self.subscription_mode = self.get("subscription_mode", default="any")
        self.show_statistics = self.get("show_statistics", default=False)
        self.statistics_interval = self.get("statistics_interval", default=30)
        self.cuckoo_warn = Cuckoo(60 * 5, log.warning)
        self.performance_warn = Cuckoo(
            self.statistics_interval, self.show_performance_statistics
        )

        self.idle_dt = deque(maxlen=1000)
        self.idle_timer = time.time()
        self.message_count = 0

        self.loop_cycle = 0
        self.queue = Queue()
        self.client = None
        self.thread = None

        if self.subscription_mode == "all":
            self.log.warning(
                "You subscribed to the ligier in 'all'-mode! "
                "If you are too slow with data processing, "
                "you will block other clients. "
                "If you don't understand this message "
                "and are running this code on a DAQ machine, "
                "consult a DAQ expert now and stop this script."
            )

        print(
            "Connecting to {0} on port {1}\n"
            "Subscribed tags: {2}\n"
            "Connection timeout: {3}s\n"
            "Maximum queue size for incoming data: {4}".format(
                self.host, self.port, self.tags, self.timeout, self.max_queue
            )
        )

        self._init_controlhost()
        self._start_thread()

    def _start_thread(self):
        log.debug("Starting and demonising thread.")
        self.thread = threading.Thread(target=self._run, args=())
        self.thread.daemon = True
        self.thread.start()

    def _init_controlhost(self):
        """Set up the controlhost connection"""
        log.debug("Connecting to JLigier")
        self.client = Client(self.host, self.port)
        self.client._connect()
        log.debug("Subscribing to tags: %s", self.tags)
        for tag in self.tags.split(","):
            self.client.subscribe(tag.strip(), mode=self.subscription_mode)
        log.debug("Controlhost initialisation done.")

    def _run(self):
        log.debug("Entering the main loop.")
        while True:
            current_qsize = self.queue.qsize()
            self.loop_cycle += 1
            self._set_idle_timer()
            try:
                prefix, data = self.client.get_message()
            except EOFError:
                log.warning("EOF from Ligier, trying again in 30 seconds...")
                time.sleep(30)
                try:
                    log.debug("Reinitialising new CH connection.")
                    self._init_controlhost()
                except socket.error:
                    log.error("Failed to connect to host.")
                continue
            except BufferError:
                log.error("Buffer error in Ligier stream, aborting...")
                break
            else:
                self._add_idle_dt()
                self.message_count += 1
                self.performance_warn()
                # log.debug("%d bytes received from network.", len(data))
            if not data:
                log.critical(
                    "No data received, connection died.\n"
                    + "Trying to reconnect in 30 seconds."
                )
                time.sleep(30)
                try:
                    log.debug("Reinitialising new CH connection.")
                    self._init_controlhost()
                except socket.error:
                    log.error("Failed to connect to host.")
                continue
            if current_qsize > self.max_queue:
                self.cuckoo_warn(
                    "Maximum queue size ({0}) reached, "
                    "dropping data.".format(self.max_queue)
                )
            else:
                self.queue.put((prefix, data))
        log.debug("Quitting the main loop.")

    def process(self, blob):
        """Wait for the next packet and put it in the blob"""
        try:
            log.debug("Waiting for queue items.")
            prefix, data = self.queue.get(timeout=self.timeout)
            log.debug("Got %d bytes from queue.", len(data))
        except Empty:
            log.warning("ControlHost timeout (%d s) reached", self.timeout)
            raise StopIteration("ControlHost timeout reached.")
        blob[self.key_for_prefix] = prefix
        blob[self.key_for_data] = data
        return blob

    def show_performance_statistics(self):
        if not self.show_statistics:
            return
        dt = np.median(self.idle_dt)
        current_qsize = self.queue.qsize()
        log_func = self.cprint
        if dt < 0 or current_qsize > 0:
            log_func = self.log.warning
        log_func(
            "Message rate: {0:.1f} Hz, median idle time per message: "
            "{1:.3f} us (current queue size: {2})".format(
                self.message_count / self.statistics_interval, dt * 1e6, current_qsize
            )
        )
        self.message_count = 0

    def _set_idle_timer(self):
        self.idle_timer = time.time()

    def _add_idle_dt(self):
        now = time.time()
        self.idle_dt.append(now - self.idle_timer)

    def finish(self):
        """Clean up the JLigier controlhost connection"""
        log.debug("Disconnecting from JLigier.")
        self.client.socket.shutdown(socket.SHUT_RDWR)
        self.client._disconnect()

    def __iter__(self):
        return self

    def __next__(self):
        return self.process(Blob())

    def next(self):
        return self.__next__()


def CHTagger(blob):
    tag = str(blob["CHPrefix"].tag)
    blob[tag] = True
    return blob



import numpy as np
import km3net_testdata
import tempfile
import h5py
import km3pipe as kp
from km3pipe.utils.h5extractf import h5extractf
from km3pipe.testing import TestCase


class TestJsireneH5File(TestCase):
    @classmethod
    def setUpClass(cls):
        filename = km3net_testdata.data_path(
            "offline/mcv5.0.DAT004340.propa.sirene.jte.jchain.aanet.4340.root"
        )
        cls.h5file = tempfile.NamedTemporaryFile()
        h5extractf(filename, outfile=cls.h5file)

    @classmethod
    def tearDownClass(cls):
        cls.h5file.close()

    def test_datasets_names_and_lengths(self):
        target = {
            "event_info": 3,
            "group_info": 3,
            "hits": 280,
            "mc_hits": 598,
            "mc_tracks": 48,
            "raw_header": 21,
            "reco": 2,
        }
        with h5py.File(self.h5file, "r") as f:
            self.assertDictEqual(target, {x: len(f[x]) for x in f})

    def test_reco_datasets_names_and_lengths(self):
        target = {"best_jmuon": 3, "tracks": 76}
        with h5py.File(self.h5file, "r") as f:
            self.assertDictEqual(target, {x: len(f["reco"][x]) for x in f["reco"]})

    def test_time_of_first_three_hits(self):
        with h5py.File(self.h5file, "r") as f:
            np.testing.assert_array_equal(
                f["hits"]["time"][:3],
                [87530475.0, 87526107.0, 87527368.0],
            )

    def test_time_of_last_three_hits(self):
        with h5py.File(self.h5file, "r") as f:
            np.testing.assert_array_equal(
                f["hits"]["time"][-3:],
                [8166085.0, 8165582.0, 8163487.0],
            )

    def test_hits_group_ids(self):
        target = [0] * 86 + [1] * 111 + [2] * 83
        with h5py.File(self.h5file, "r") as f:
            np.testing.assert_array_equal(
                f["hits"]["group_id"],
                target,
            )

    def test_hits_dtype_names(self):
        target = (
            "channel_id",
            "dom_id",
            "time",
            "tot",
            "triggered",
            "pos_x",
            "pos_y",
            "pos_z",
            "dir_x",
            "dir_y",
            "dir_z",
            "tdc",
            "group_id",
        )
        with h5py.File(self.h5file, "r") as f:
            self.assertTupleEqual(
                f["hits"].dtype.names,
                target,
            )

    def test_h5_file_can_be_opened_with_hdf5pump_and_keys_are_correct(self):
        pump = kp.io.HDF5Pump(filename=self.h5file.name)
        blob = pump[1]
        target = {
            "BestJmuon",
            "EventInfo",
            "GroupInfo",
            "Header",
            "Hits",
            "McHits",
            "McTracks",
            "RawHeader",
            "Tracks",
        }
        self.assertSetEqual(target, set(blob.keys()))



# Filename: time.py
# pylint: disable=C0103
"""
Manipulating time and so...

"""
from datetime import datetime
import numpy as np
import time
from timeit import default_timer as timer

__author__ = "Tamas Gal and Moritz Lotze"
__copyright__ = "Copyright 2016, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal and Moritz Lotze"
__email__ = "tgal@km3net.de"
__status__ = "Development"


def total_seconds(td):
    """Convert the timedelta to seconds."""
    return td.total_seconds()


class Timer(object):
    """A very simple, accurate and easy to use timer context"""

    def __init__(self, message="It", precision=3, callback=print):
        self.message = message
        self.precision = precision
        self.callback = callback

    def __enter__(self):
        self.start()

    def __exit__(self, exit_type, value, traceback):
        self.stop()

    def start(self):
        self.__start = timer()
        self.__start_cpu = time.process_time()

    def stop(self):
        self.__finish = timer()
        self.__finish_cpu = time.process_time()
        if self.callback is not None:
            self.log()
        return self.seconds

    @property
    def seconds(self):
        return self.__finish - self.__start

    @property
    def cpu_seconds(self):
        return self.__finish_cpu - self.__start_cpu

    def log(self):
        self.callback(
            "{0} took {1:.{3}f}s (CPU {2:.{3}f}s).".format(
                self.message, self.seconds, self.cpu_seconds, self.precision
            )
        )


class Cuckoo(object):
    "A timed callback caller, which only executes once in a given interval."

    def __init__(self, interval=0, callback=print):
        "Setup with interval in seconds and a callback function"
        self.interval = interval
        self.callback = callback
        self.timestamp = None

    def msg(self, *args, **kwargs):
        "Only execute callback when interval is reached."
        if self.timestamp is None or self._interval_reached():
            self.callback(*args, **kwargs)
            self.reset()

    def reset(self):
        "Reset the timestamp"
        self.timestamp = datetime.now()

    def _interval_reached(self):
        "Check if defined interval is reached"
        return total_seconds(datetime.now() - self.timestamp) > self.interval

    def __call__(self, *args, **kwargs):
        "Run the msg function when called directly."
        self.msg(*args, **kwargs)


def tai_timestamp():
    """Return current TAI timestamp."""
    timestamp = time.time()
    date = datetime.utcfromtimestamp(timestamp)
    if date.year < 1972:
        return timestamp
    offset = 10 + timestamp
    leap_seconds = [
        (1972, 1, 1),
        (1972, 7, 1),
        (1973, 1, 1),
        (1974, 1, 1),
        (1975, 1, 1),
        (1976, 1, 1),
        (1977, 1, 1),
        (1978, 1, 1),
        (1979, 1, 1),
        (1980, 1, 1),
        (1981, 7, 1),
        (1982, 7, 1),
        (1983, 7, 1),
        (1985, 7, 1),
        (1988, 1, 1),
        (1990, 1, 1),
        (1991, 1, 1),
        (1992, 7, 1),
        (1993, 7, 1),
        (1994, 7, 1),
        (1996, 1, 1),
        (1997, 7, 1),
        (1999, 1, 1),
        (2006, 1, 1),
        (2009, 1, 1),
        (2012, 7, 1),
        (2015, 7, 1),
        (2017, 1, 1),
    ]
    for idx, leap_date in enumerate(leap_seconds):
        if leap_date >= (date.year, date.month, date.day):
            return idx - 1 + offset
    return len(leap_seconds) - 1 + offset


def np_to_datetime(intime):
    """Convert numpy/pandas datetime64 to list[datetime]."""
    nptime = np.atleast_1d(intime)
    np_corr = (nptime - np.datetime64("1970-01-01T00:00:00")) / np.timedelta64(1, "s")
    return [datetime.utcfromtimestamp(t) for t in np_corr]



# Filename: test_online.py
# pylint: disable=locally-disabled,C0111,R0904,C0301,C0103,W0212
from km3pipe.testing import TestCase, patch, skip, data_path
import sys
import numpy as np
from km3pipe.io.online import EventPump, TimeslicePump
from km3pipe.io.hdf5 import HDF5Pump
from os.path import join, dirname, abspath

__author__ = "Tamas Gal"
__copyright__ = "Copyright 2018, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal, Johannes Schumann"
__email__ = "tgal@km3net.de"
__status__ = "Development"

ONLINE_TEST_FILE = data_path("online/events_sample.root")
ONLINE_TEST_FILE_H5 = data_path("online/events_sample.h5")


class TestEventPump(TestCase):
    def setUp(self):
        self.jpp_pump = EventPump(filename=ONLINE_TEST_FILE)
        self.h5_pump = HDF5Pump(filename=ONLINE_TEST_FILE_H5)

    def tearDown(self):
        self.jpp_pump.finish()
        self.h5_pump.finish()

    def test_event_info(self):
        n = self.jpp_pump.n_events
        for i in range(n):
            h5blob = self.h5_pump[i]
            jppblob = self.jpp_pump[i]
            for k in h5blob["EventInfo"].dtype.names:
                if k not in jppblob["EventInfo"].dtype.names:
                    continue
                ref_value = h5blob["EventInfo"][k][0]
                test_value = jppblob["EventInfo"][k][0]
                if isinstance(ref_value, float):
                    if np.isnan(ref_value):
                        assert np.isnan(test_value) == np.isnan(ref_value)
                    else:
                        self.assertAlmostEqual(test_value, ref_value)
                else:
                    assert ref_value == test_value

    @skip
    def test_hit_info(self):
        n = self.jpp_pump.n_events
        for i in range(n):
            h5blob = self.h5_pump[i]
            jppblob = self.jpp_pump[i]
            for k in h5blob["Hits"].dtype.names:
                ref_value = h5blob["Hits"][k][0]
                test_value = jppblob["Hits"][k][0]
                if isinstance(ref_value, float):
                    if np.isnan(ref_value):
                        assert np.isnan(test_value)
                    else:
                        self.assertAlmostEqual(test_value, ref_value)
                else:
                    assert ref_value == test_value


class TestTimeslicePump(TestCase):
    def setUp(self):
        self.jpp_fname = ONLINE_TEST_FILE
        self.timeslice_pump = TimeslicePump(filename=self.jpp_fname, stream="L1")

    def test_timeslice_info(self):
        blob = self.timeslice_pump[0]
        tsinfo = blob["TimesliceInfo"]
        assert tsinfo.frame_index == 366
        assert tsinfo.slice_id == 0
        assert tsinfo.timestamp == 1575979236
        assert tsinfo.nanoseconds == 600000000
        assert tsinfo.n_frames == 69

    def test_timeslice_hits(self):
        blob = self.timeslice_pump[0]
        l1hits = blob["L1Hits"]
        counts = [
            1421,
            1272,
            1533,
            1209,
            1311,
            1450,
            1102,
            1494,
            1410,
            1390,
            1225,
            1419,
            1474,
            1452,
            1589,
            1362,
            1577,
            1598,
            1891,
            1584,
            1571,
            1523,
            1696,
            1719,
            1540,
            1578,
            1656,
            1569,
            1422,
            1401,
            1262,
        ]
        tot = [
            216,
            681,
            1222,
            1580,
            1484,
            1221,
            940,
            887,
            908,
            950,
            844,
            736,
            717,
            661,
            752,
            716,
            791,
            946,
            1089,
            1220,
            1484,
            1786,
            2265,
            2539,
            2932,
            3132,
            3201,
            2901,
            2403,
            1654,
            1056,
            566,
            297,
            167,
            94,
            74,
            82,
            60,
            50,
            37,
            30,
            24,
            24,
            20,
            19,
            12,
            15,
            15,
            13,
            8,
            5,
            6,
            9,
            9,
            7,
            9,
            11,
            6,
            5,
            10,
            7,
            4,
            7,
            2,
            7,
            1,
            4,
            5,
            5,
            3,
            3,
            3,
            2,
            3,
            2,
            3,
            1,
            4,
            3,
            4,
            1,
            1,
            2,
            2,
            1,
            2,
            3,
            1,
            1,
            1,
            2,
            1,
            1,
            1,
            1,
            1,
            1,
            1,
            1,
            1,
            1,
            1,
            1,
        ]
        dom_id = [
            774,
            244,
            602,
            935,
            533,
            878,
            862,
            699,
            760,
            1058,
            625,
            641,
            621,
            679,
            912,
            461,
            925,
            633,
            754,
            701,
            767,
            784,
            812,
            299,
            605,
            1010,
            226,
            892,
            785,
            864,
            707,
            540,
            749,
            772,
            172,
            748,
            773,
            585,
            278,
            656,
            852,
            685,
            1041,
            473,
            639,
            695,
            837,
            830,
            656,
            676,
            486,
            647,
            533,
            716,
            649,
            599,
            502,
            271,
            629,
            769,
            613,
            784,
            705,
            600,
            984,
            682,
            243,
            583,
        ]
        self.assertListEqual(
            counts, np.unique(l1hits.channel_id, return_counts=True)[1].tolist()
        )
        self.assertListEqual(tot, np.unique(l1hits.tot, return_counts=True)[1].tolist())
        self.assertListEqual(
            dom_id, np.unique(l1hits.dom_id, return_counts=True)[1].tolist()
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# vim: ts=4 sw=4 et
"""
=====================
Trigger Contributions
=====================

Shows the median trigger contribution for each DOM.
This script can be used to easily identify DOMs in a run, which are out
of sync.

"""

from collections import defaultdict
import sys
import numpy as np
import km3pipe as kp
from km3modules.common import StatusBar

if len(sys.argv) == 2:
    filename = sys.argv[1]
else:
    raise SystemExit("Usage: trigger_contribution.py FILENAME")

det = kp.hardware.Detector(det_id=29)
log = kp.logger.get_logger("TriggerContribution")


class TriggerContributionCalculator(kp.Module):
    """Shows the mean trigger contribution for each DOM"""

    def configure(self):
        self.dus = self.get("dus")  # only select DOMs on these DUs
        self.trigger_contributions = defaultdict(list)
        self.n_events = 0

    def process(self, blob):
        hits = blob["Hits"].triggered_rows
        n_hits = len(hits)
        dom_ids = np.unique(hits.dom_id)
        for dom_id in dom_ids:
            trigger_contribution = np.sum(hits.dom_id == dom_id) / n_hits
            self.trigger_contributions[dom_id].append(trigger_contribution)
        self.n_events += 1
        return blob

    def finish(self):
        print(
            "{}\n{:>12}  {:>4} {:>4}  {:>12}\n{}".format(
                "=" * 42, "DOM ID", "du", "floor", "trig. contr.", "-" * 42
            )
        )
        summary = []
        for dom_id, trigger_contribution in self.trigger_contributions.items():
            du, floor = omkey(dom_id)
            mean_tc = np.sum(trigger_contribution) / self.n_events
            summary.append(((du, floor), dom_id, mean_tc))
        for (du, floor), dom_id, mean_tc in sorted(summary):
            print(
                "{:>12}  {:>4} {:>4}  {:>12.2f}%".format(
                    dom_id, du, floor, mean_tc * 100
                )
            )

        dom_ids = set(det.doms.keys())
        if self.dus is not None:
            log.warning(
                "Showing only DOMs which are on the following DUs: {}".format(
                    ", ".join(str(du) for du in self.dus)
                )
            )
            dom_ids = set(d for d in dom_ids if det.doms[d][0] in self.dus)

        inactive_doms = []
        for dom_id in set(dom_ids) - set(self.trigger_contributions.keys()):
            inactive_doms.append(dom_id)
        if inactive_doms:
            print("The following DOMs were inactive:")
            for dom_id in inactive_doms:
                print("{}_(DU{}-{})".format(dom_id, *omkey(dom_id)), end=" ")


def omkey(dom_id):
    """Returns (du, floor) for given DOM ID"""
    return det.doms[dom_id][0:2]


pipe = kp.Pipeline()
pipe.attach(kp.io.jpp.EventPump, filename=filename)
pipe.attach(StatusBar, every=5000)
pipe.attach(TriggerContributionCalculator, dus=[2])
pipe.drain()



# Filename: __init__.py
"""
The extemporary KM3NeT analysis framework.

"""
from pkg_resources import get_distribution, DistributionNotFound

try:
    version = get_distribution(__name__).version
except DistributionNotFound:
    version = "unknown version"

try:
    __KM3PIPE_SETUP__
except NameError:
    __KM3PIPE_SETUP__ = False

if not __KM3PIPE_SETUP__:
    from . import logger  # noqa
    from .dataclasses import Table, NDArray  # noqa
    from . import dataclasses  # noqa
    from . import calib  # noqa
    from . import cmd  # noqa
    from . import constants  # noqa
    from . import controlhost  # noqa
    from . import hardware  # noqa
    from . import io  # noqa
    from . import math  # noqa
    from . import mc  # noqa
    from . import physics
    from . import shell  # noqa
    from . import style  # noqa
    from . import sys  # noqa

    # from . import testing     # noqa
    from . import time  # noqa
    from . import tools  # noqa

    from thepipe import (
        Pipeline,
        Module,
        Blob,
        Provenance,
    )  # reexport the provenance handler

__author__ = "Tamas Gal and Moritz Lotze"
__copyright__ = "Copyright 2016, Tamas Gal and the KM3NeT collaboration."
__credits__ = ["Thomas Heid"]
__license__ = "MIT"
__version__ = version
__maintainer__ = "Tamas Gal and Moritz Lotze"
__email__ = "tgal@km3net.de"
__status__ = "Development"



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# vim:set ts=4 sts=4 sw=4 et:
"""
Functions and modules to work with hits.
"""

from numba import njit
import numpy as np

import km3pipe as kp

log = kp.logger.get_logger(__name__)

__author__ = "Tamas Gal"


@njit
def count_multiplicities(times, tmax=20):
    """Calculate an array of multiplicities and corresponding coincidence IDs

    Note that this algorithm does not take care about DOM IDs, so it has to
    be fed with DOM hits.

    Parameters
    ----------
    times: array[float], shape=(n,)
        Hit times for n hits
    dt: int [default: 20]
        Time window of a coincidence

    Returns
    -------
    (array[int]), array[int]), shape=(n,)

    """
    n = times.shape[0]
    mtp = np.ones(n, dtype=np.int64)  # multiplicities
    cid = np.zeros(n, dtype=np.int64)  # coincidence id
    idx0 = 0
    _mtp = 1
    _cid = 0
    t0 = times[idx0]
    for i in range(1, n):
        dt = times[i] - t0
        if dt > tmax:
            mtp[idx0:i] = _mtp
            cid[idx0:i] = _cid
            _mtp = 0
            _cid += 1
            idx0 = i
            t0 = times[i]
        _mtp += 1
        if i == n - 1:
            mtp[idx0:] = _mtp
            cid[idx0:] = _cid
            break

    return mtp, cid



# Filename: test_hits.py
# -*- coding: utf-8 -*-
# vim:set ts=4 sts=4 sw=4 et:
"""
Tests for Hit functions and Modules.
"""

import numpy as np
from numpy.testing import assert_array_equal, assert_allclose

from km3modules.hits import count_multiplicities

from km3pipe import Table, Blob, Pipeline, Module
from km3pipe.testing import TestCase


class TestMultiplicityCounter(TestCase):
    def test_count_multiplicities(self):
        times = np.array([1, 10, 20, 22, 30, 34, 40, 50])
        mtps, coinc_ids = count_multiplicities(times, tmax=20)
        self.assertListEqual([3, 3, 3, 4, 4, 4, 4, 1], list(mtps))
        self.assertListEqual([0, 0, 0, 1, 1, 1, 1, 2], list(coinc_ids))

    def test_count_other_multiplicities(self):
        times = np.array([1, 10, 20, 22, 30, 34, 40, 50])
        mtps, coinc_ids = count_multiplicities(times, tmax=10)
        self.assertListEqual([2, 2, 3, 3, 3, 2, 2, 1], list(mtps))
        self.assertListEqual([0, 0, 1, 1, 1, 2, 2, 3], list(coinc_ids))



# Filename: test_calib.py
# pylint: disable=C0111,E1003,R0904,C0103,R0201,C0102
from os.path import dirname, join
import functools
import operator
import shutil
import sys
import tempfile

import km3io
from thepipe import Module, Pipeline
import km3pipe as kp
from km3pipe.dataclasses import Table
from km3pipe.hardware import Detector
from km3pipe.io.hdf5 import HDF5Sink
from km3pipe.testing import TestCase, MagicMock, patch, skip, skipif, data_path
from km3pipe.calib import Calibration, CalibrationService, slew

from .test_hardware import EXAMPLE_DETX

import numpy as np
import tables as tb

__author__ = "Tamas Gal"
__copyright__ = "Copyright 2016, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal"
__email__ = "tgal@km3net.de"
__status__ = "Development"


class TestCalibration(TestCase):
    """Tests for the Calibration class"""

    def test_init_with_wrong_file_extension(self):
        with self.assertRaises(NotImplementedError):
            Calibration(filename="foo")

    @patch("km3pipe.calib.Detector")
    def test_init_with_filename(self, mock_detector):
        Calibration(filename="foo.detx")
        mock_detector.assert_called_with(filename="foo.detx")

    @patch("km3pipe.calib.Detector")
    def test_init_with_det_id(self, mock_detector):
        Calibration(det_id=1)
        mock_detector.assert_called_with(t0set=None, calibset=None, det_id=1)
        Calibration(det_id=1, calibset=2, t0set=3)
        mock_detector.assert_called_with(t0set=3, calibset=2, det_id=1)

    def test_init_with_detector(self):
        det = Detector(data_path("detx/detx_v1.detx"))
        Calibration(detector=det)

    def test_apply_to_hits_with_pmt_id_aka_mc_hits(self):
        calib = Calibration(filename=data_path("detx/detx_v1.detx"))

        hits = Table({"pmt_id": [1, 2, 1], "time": [10.1, 11.2, 12.3]})

        chits = calib.apply(hits, correct_slewing=False)

        assert len(hits) == len(chits)

        a_hit = chits[0]
        self.assertAlmostEqual(1.1, a_hit.pos_x)
        self.assertAlmostEqual(10, a_hit.t0)
        self.assertAlmostEqual(10.1, a_hit.time)  # t0 should not bei applied

        a_hit = chits[1]
        self.assertAlmostEqual(1.4, a_hit.pos_x)
        self.assertAlmostEqual(20, a_hit.t0)
        self.assertAlmostEqual(11.2, a_hit.time)  # t0 should not be applied

    def test_apply_to_hits_with_pmt_id_aka_mc_hits_from_km3io(self):
        calib = Calibration(filename=data_path("detx/KM3NeT_-00000001_20171212.detx"))
        f = km3io.OfflineReader(
            data_path(
                "offline/mcv6.gsg_nue-CCHEDIS_1e4-1e6GeV.sirene.jte.jchain.aanet.1.root"
            )
        )

        for event in f:
            chits = calib.apply(event.mc_hits)
            assert 840 == len(chits.t0)
            assert np.allclose([3, 26, 24, 4, 23, 25], chits.channel_id[:6])
            assert np.allclose([3401, 3401, 3406, 3411, 5501, 5501], chits.dom_id[:6])
            assert np.allclose([1, 1, 6, 11, 1, 1], chits.floor[:6])
            assert np.allclose([34, 34, 34, 34, 55, 55], chits.du[:6])
            assert np.allclose(
                [
                    1679.18706571,
                    1827.14262054,
                    1926.71722628,
                    2433.83097585,
                    1408.35942832,
                    1296.51397496,
                ],
                chits.time[:6],
            )
            assert np.allclose(
                [2.034, 1.847, 1.938, 2.082, -54.96, -55.034], chits.pos_x[:6]
            )
            assert np.allclose(
                [-233.415, -233.303, -233.355, -233.333, -341.346, -341.303],
                chits.pos_y[:6],
            )
            assert np.allclose(
                [65.059, 64.83, 244.83, 425.111, 64.941, 64.83], chits.pos_z[:6]
            )
            assert np.allclose([4, 4, 4, 26, 4, 4], f.mc_hits.origin[0][:6].tolist())
            assert np.allclose(
                [36835, 36881, 37187, 37457, 60311, 60315],
                f.mc_hits.pmt_id[0][:6].tolist(),
            )
            break

    def test_dus(self):
        calib = Calibration(filename=data_path("detx/detx_v1.detx"))

        hits = Table(
            {"dom_id": [2, 6, 3], "channel_id": [0, 1, 2], "time": [10.1, 11.2, 12.3]}
        )

        dus = calib.dus(hits)
        assert np.allclose([1, 2, 1], dus)

    def test_floors(self):
        calib = Calibration(filename=data_path("detx/detx_v1.detx"))

        hits = Table(
            {"dom_id": [2, 6, 3], "channel_id": [0, 1, 2], "time": [10.1, 11.2, 12.3]}
        )

        floors = calib.floors(hits)
        assert np.allclose([2, 3, 3], floors)

    def test_apply_to_hits_with_dom_id_and_channel_id(self):
        calib = Calibration(filename=data_path("detx/detx_v1.detx"))

        hits = Table(
            {"dom_id": [2, 3, 3], "channel_id": [0, 1, 2], "time": [10.1, 11.2, 12.3]}
        )

        chits = calib.apply(hits, correct_slewing=False)

        assert len(hits) == len(chits)

        a_hit = chits[0]
        self.assertAlmostEqual(2.1, a_hit.pos_x)
        self.assertAlmostEqual(40, a_hit.t0)
        t0 = a_hit.t0
        self.assertAlmostEqual(10.1 + t0, a_hit.time)

        a_hit = chits[1]
        self.assertAlmostEqual(3.4, a_hit.pos_x)
        self.assertAlmostEqual(80, a_hit.t0)
        t0 = a_hit.t0
        self.assertAlmostEqual(11.2 + t0, a_hit.time)

    def test_assert_apply_adds_dom_id_and_channel_id_to_mc_hits(self):
        calib = Calibration(filename=data_path("detx/detx_v1.detx"))
        hits = Table({"pmt_id": [1, 2, 1], "time": [10.1, 11.2, 12.3]})
        chits = calib.apply(hits)
        self.assertListEqual([1, 1, 1], list(chits.dom_id))
        self.assertListEqual([0, 1, 0], list(chits.channel_id))

    def test_assert_apply_adds_pmt_id_to_hits(self):
        calib = Calibration(filename=data_path("detx/detx_v1.detx"))
        hits = Table(
            {"dom_id": [2, 3, 3], "channel_id": [0, 1, 2], "time": [10.1, 11.2, 12.3]}
        )
        chits = calib.apply(hits, correct_slewing=False)
        self.assertListEqual([4, 8, 9], list(chits.pmt_id))

    def test_apply_to_hits_with_pmt_id_with_wrong_calib_raises(self):
        calib = Calibration(filename=data_path("detx/detx_v1.detx"))

        hits = Table({"pmt_id": [999], "time": [10.1]})

        with self.assertRaises(KeyError):
            calib.apply(hits, correct_slewing=False)

    def test_apply_to_hits_with_dom_id_and_channel_id_with_wrong_calib_raises(self):
        calib = Calibration(filename=data_path("detx/detx_v1.detx"))

        hits = Table({"dom_id": [999], "channel_id": [0], "time": [10.1]})

        with self.assertRaises(KeyError):
            calib.apply(hits, correct_slewing=False)

    def test_apply_to_hits_from_km3io(self):
        calib = Calibration(filename=data_path("detx/km3net_offline.detx"))
        hits = km3io.OfflineReader(data_path("offline/km3net_offline.root"))[0].hits

        chits = calib.apply(hits)
        assert 176 == len(chits.t0)
        assert np.allclose([207747.825, 207745.656, 207743.836], chits.t0.tolist()[:3])

        chits = calib.apply(hits[:3])
        assert 3 == len(chits.t0)
        assert np.allclose([207747.825, 207745.656, 207743.836], chits.t0.tolist()[:3])

    def test_apply_to_hits_from_km3io_iterator(self):
        calib = Calibration(filename=data_path("detx/km3net_offline.detx"))
        f = km3io.OfflineReader(data_path("offline/km3net_offline.root"))

        for event in f:
            chits = calib.apply(event.hits)
            assert 176 == len(chits.t0)
            assert np.allclose(
                [207747.825, 207745.656, 207743.836], chits.t0.tolist()[:3]
            )
            break

    def test_time_slewing_correction(self):
        calib = Calibration(filename=data_path("detx/detx_v1.detx"))

        hits = Table(
            {
                "dom_id": [2, 3, 3],
                "channel_id": [0, 1, 2],
                "time": [10.1, 11.2, 12.3],
                "tot": [0, 10, 255],
            }
        )

        chits = calib.apply(hits)  #  correct_slewing=True is default

        assert len(hits) == len(chits)

        a_hit = chits[0]
        self.assertAlmostEqual(10.1 + a_hit.t0 - slew(a_hit.tot), a_hit.time)

        a_hit = chits[1]
        self.assertAlmostEqual(11.2 + a_hit.t0 - slew(a_hit.tot), a_hit.time)

        a_hit = chits[2]
        self.assertAlmostEqual(12.3 + a_hit.t0 - slew(a_hit.tot), a_hit.time)

    def test_apply_to_timeslice_hits(self):
        tshits = Table.from_template(
            {
                "channel_id": [0, 1, 2],
                "dom_id": [2, 3, 3],
                "time": [10.1, 11.2, 12.3],
                "tot": np.ones(3, dtype=float),
                "group_id": 0,
            },
            "TimesliceHits",
        )
        calib = Calibration(filename=data_path("detx/detx_v1.detx"))
        c_tshits = calib.apply(tshits, correct_slewing=False)
        assert len(c_tshits) == len(tshits)
        assert np.allclose([40, 80, 90], c_tshits.t0)
        # TimesliceHits is using int4 for times, so it's truncated when we pass in float64
        assert np.allclose([50.1, 91.2, 102.3], c_tshits.time, atol=0.1)

    def test_apply_without_affecting_primary_hit_table(self):
        calib = Calibration(filename=data_path("detx/detx_v1.detx"))
        hits = Table({"pmt_id": [1, 2, 1], "time": [10.1, 11.2, 12.3]})
        hits_compare = hits.copy()
        calib.apply(hits, correct_slewing=False)

        for t_primary, t_calib in zip(hits_compare, hits):
            self.assertAlmostEqual(t_primary, t_calib)

    def test_calibration_in_pipeline(self):
        class DummyPump(kp.Module):
            def configure(self):
                self.index = 0

            def process(self, blob):
                self.index += 1
                mc_hits = Table({"pmt_id": [1, 2, 1], "time": [10.1, 11.2, 12.3]})
                hits = Table(
                    {
                        "dom_id": [2, 3, 3],
                        "channel_id": [0, 1, 2],
                        "time": [10.1, 11.2, 12.3],
                        "tot": [0, 10, 255],
                    }
                )

                blob["Hits"] = hits
                blob["McHits"] = mc_hits
                return blob

        _self = self

        class Observer(kp.Module):
            def process(self, blob):
                assert "Hits" in blob
                assert "McHits" in blob
                assert "CalibHits" in blob
                assert "CalibMcHits" in blob
                assert not hasattr(blob["Hits"], "pmt_id")
                assert hasattr(blob["CalibHits"], "pmt_id")
                assert not hasattr(blob["McHits"], "dom_id")
                assert hasattr(blob["CalibHits"], "dom_id")
                assert np.allclose([10.1, 11.2, 12.3], blob["Hits"].time)
                assert np.allclose([42.09, 87.31, 111.34], blob["CalibHits"].time)
                assert np.allclose(blob["McHits"].time, blob["CalibMcHits"].time)
                return blob

        pipe = kp.Pipeline()
        pipe.attach(DummyPump)
        pipe.attach(Calibration, filename=data_path("detx/detx_v1.detx"))
        pipe.attach(Observer)
        pipe.drain(3)


class TestCalibrationService(TestCase):
    def test_apply_to_hits_with_dom_id_and_channel_id(self):

        hits = Table(
            {
                "dom_id": [2, 3, 3],
                "channel_id": [0, 1, 2],
                "time": [10.1, 11.2, 12.3],
                "tot": [23, 105, 231],
            }
        )

        tester = self

        class HitCalibrator(Module):
            def process(self, blob):
                chits = self.services["calibrate"](hits)

                assert len(hits) == len(chits)

                a_hit = chits[0]
                tester.assertAlmostEqual(2.1, a_hit.pos_x)
                tester.assertAlmostEqual(40, a_hit.t0)
                t0 = a_hit.t0
                tester.assertAlmostEqual(10.1 + t0 - slew(a_hit.tot), a_hit.time)

                a_hit = chits[1]
                tester.assertAlmostEqual(3.4, a_hit.pos_x)
                tester.assertAlmostEqual(80, a_hit.t0)
                t0 = a_hit.t0
                tester.assertAlmostEqual(11.2 + t0 - slew(a_hit.tot), a_hit.time)
                return blob

        pipe = Pipeline()
        pipe.attach(CalibrationService, filename=data_path("detx/detx_v1.detx"))
        pipe.attach(HitCalibrator)
        pipe.drain(1)

    def test_apply_to_hits_with_dom_id_and_channel_id_without_slewing(self):

        hits = Table(
            {
                "dom_id": [2, 3, 3],
                "channel_id": [0, 1, 2],
                "time": [10.1, 11.2, 12.3],
                "tot": [23, 105, 231],
            }
        )

        tester = self

        class HitCalibrator(Module):
            def process(self, blob):
                chits = self.services["calibrate"](hits, correct_slewing=False)

                assert len(hits) == len(chits)

                a_hit = chits[0]
                tester.assertAlmostEqual(2.1, a_hit.pos_x)
                tester.assertAlmostEqual(40, a_hit.t0)
                t0 = a_hit.t0
                tester.assertAlmostEqual(10.1 + t0, a_hit.time)

                a_hit = chits[1]
                tester.assertAlmostEqual(3.4, a_hit.pos_x)
                tester.assertAlmostEqual(80, a_hit.t0)
                t0 = a_hit.t0
                tester.assertAlmostEqual(11.2 + t0, a_hit.time)
                return blob

        pipe = Pipeline()
        pipe.attach(CalibrationService, filename=data_path("detx/detx_v1.detx"))
        pipe.attach(HitCalibrator)
        pipe.drain(1)

    def test_correct_slewing(self):

        hits = Table(
            {
                "dom_id": [2, 3, 3],
                "channel_id": [0, 1, 2],
                "time": [10.1, 11.2, 12.3],
                "tot": [0, 10, 255],
            }
        )

        tester = self

        class HitCalibrator(Module):
            def process(self, blob):
                self.services["correct_slewing"](hits)

                a_hit = hits[0]
                tester.assertAlmostEqual(10.1 - slew(a_hit.tot), a_hit.time)

                a_hit = hits[1]
                tester.assertAlmostEqual(11.2 - slew(a_hit.tot), a_hit.time)
                return blob

        pipe = Pipeline()
        pipe.attach(CalibrationService, filename=data_path("detx/detx_v1.detx"))
        pipe.attach(HitCalibrator)
        pipe.drain(1)

    def test_provided_detector_data(self):
        class DetectorReader(Module):
            def process(self, blob):
                assert "get_detector" in self.services
                det = self.services["get_detector"]()
                assert isinstance(det, Detector)

        pipe = Pipeline()
        pipe.attach(CalibrationService, filename=data_path("detx/detx_v1.detx"))
        pipe.attach(DetectorReader)
        pipe.drain(1)


class TestSlew(TestCase):
    def test_slew(self):
        self.assertAlmostEqual(8.01, slew(0))
        self.assertAlmostEqual(0.60, slew(23))
        self.assertAlmostEqual(-9.04, slew(255))

    def test_slew_vectorised(self):
        assert np.allclose([8.01, 0.60, -9.04], slew(np.array([0, 23, 255])))
